'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var classTransformer = require('class-transformer');
var entity = require('../../dist/entity-60992a93.cjs.js');
var storage_js = require('class-transformer/cjs/storage.js');
var get = require('lodash/get');
var set$1 = require('lodash/set');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var get__default = /*#__PURE__*/_interopDefault(get);
var set__default = /*#__PURE__*/_interopDefault(set$1);

function _initializerDefineProperty(target, property, descriptor, context) {
  if (!descriptor) return;
  Object.defineProperty(target, property, {
    enumerable: descriptor.enumerable,
    configurable: descriptor.configurable,
    writable: descriptor.writable,
    value: descriptor.initializer ? descriptor.initializer.call(context) : void 0
  });
}

function _applyDecoratedDescriptor(target, property, decorators, descriptor, context) {
  var desc = {};
  Object.keys(descriptor).forEach(function (key) {
    desc[key] = descriptor[key];
  });
  desc.enumerable = !!desc.enumerable;
  desc.configurable = !!desc.configurable;
  if ('value' in desc || desc.initializer) {
    desc.writable = true;
  }
  desc = decorators.slice().reverse().reduce(function (desc, decorator) {
    return decorator(target, property, desc) || desc;
  }, desc);
  if (context && desc.initializer !== void 0) {
    desc.value = desc.initializer ? desc.initializer.call(context) : void 0;
    desc.initializer = undefined;
  }
  if (desc.initializer === void 0) {
    Object.defineProperty(target, property, desc);
    desc = null;
  }
  return desc;
}

var _dec, _class, _descriptor;
let WithMeta = (_dec = classTransformer.Expose(), (_class = class WithMeta {
  constructor() {
    _initializerDefineProperty(this, "_meta", _descriptor, this);
  }
}, (_descriptor = _applyDecoratedDescriptor(_class.prototype, "_meta", [_dec], {
  configurable: true,
  enumerable: true,
  writable: true,
  initializer: null
})), _class));

class Entity {
  constructor(params) {
    this.type = void 0;
    this.id = void 0;
    this.data = void 0;
    this.path = void 0;
    this.meta = void 0;
    this.id = params.id;
    this.type = params.type;
    this.meta = params.meta;
  }
  get entityTypeId() {
    return this.type.split("--")[0];
  }
  get bundleId() {
    return this.type.split("--")[1];
  }
}

function notEmpty(value) {
  return value !== null && value !== undefined;
}

class EntityResolver {
  constructor(httpClient, options) {
    this.httpClient = void 0;
    this.entityMap = new Map();
    this.options = void 0;
    this.options = options;
    this.httpClient = httpClient;
  }
  async execute({
    path,
    data,
    target
  }) {
    const targetEntities = Reflect.getMetadata(entity.ENTITY_KEY, target);
    if (!targetEntities) return this.entityMap;
    const targetExposedProps = storage_js.defaultMetadataStorage.getExposedMetadatas(target);
    const exposedEntityProps = targetExposedProps.filter(({
      propertyName
    }) => targetEntities.has(propertyName ?? ""));
    const incomingPath = `${path}${path?.length ? "." : ""}`;
    for (const {
      propertyName,
      options
    } of exposedEntityProps) {
      const dataAccessorName = options.name || propertyName;
      if (!dataAccessorName) throw new Error("No data accessor property. Not possible to read entities from resolved data");
      const dataResource = get__default["default"](data, dataAccessorName);
      if (propertyName) {
        const {
          options,
          reflectedType,
          typeFunction
        } = storage_js.defaultMetadataStorage.findTypeMetadata(target, propertyName);
        const type = typeFunction();
        const subTypes = options.discriminator?.subTypes;
        const entityDecoratorOptions = targetEntities.get(propertyName ?? "");
        const params = {
          subTypes,
          type,
          incomingPath,
          dataAccessorName
        };
        if (reflectedType.name === "Array" && Array.isArray(dataResource)) {
          await this.arrayTypeResolver(params, dataResource, data, entityDecoratorOptions);
        } else if (reflectedType.name === "Object" && !Array.isArray(dataResource)) {
          const shouldResolve = await this.conditionalResolver({
            entityDecoratorOptions,
            meta: dataResource.meta,
            parentData: data,
            dataAccessorName
          });
          if (!shouldResolve) continue;
          await this.objectTypeResolver(params, dataResource);
        }
      }
    }
    return this.entityMap;
  }
  async conditionalResolver({
    entityDecoratorOptions,
    dataAccessorName,
    parentData,
    meta
  }) {
    if (entityDecoratorOptions?.conditionalResolver) {
      const hasIndexMatch = dataAccessorName.match(/\[(.*?)]/);
      const sanitizedMatch = hasIndexMatch?.[0].replace("[", "").replace("]", "");
      const idx = sanitizedMatch ? parseInt(sanitizedMatch) : undefined;
      const shouldResolve = entityDecoratorOptions.conditionalResolver(this.options?.context ?? {}, meta || {}, idx);
      if (!shouldResolve) {
        set__default["default"](parentData, dataAccessorName, idx !== undefined ? {} : undefined);
      }
      return shouldResolve;
    }
    return true;
  }
  async objectTypeResolver({
    incomingPath,
    type,
    subTypes,
    dataAccessorName
  }, dataResource) {
    const entity = this.createEntity(dataResource);
    entity.path = `${incomingPath}${dataAccessorName}`;
    entity.data = await this.httpClient.get(`/${entity.entityTypeId}/${entity.bundleId}/${entity.id}`);
    this.entityMap.set(entity.path, entity);
    if (!entity.data) throw new Error("Entity has no resolved data");
    if (subTypes) {
      const correspondingSubType = subTypes.find(subType => subType.name === entity.type);
      if (!correspondingSubType) throw new Error("Error retrieving subtype for object entity union");
      await this.execute({
        target: correspondingSubType.value,
        data: entity.data,
        path: entity.path
      });
    } else {
      await this.execute({
        target: type,
        data: entity.data,
        path: entity.path
      });
    }
  }
  async arrayTypeResolver({
    dataAccessorName,
    type,
    subTypes,
    incomingPath
  }, dataResource, parentData, entityDecoratorOptions) {
    const baseParams = {
      dataResourceList: dataResource,
      incomingPath,
      dataAccessorName,
      entityDecoratorOptions,
      parentData
    };
    if (subTypes) {
      const orderedSubTypes = dataResource.map(resourceEntity => subTypes.find(type => type.name === resourceEntity.type)?.value).filter(notEmpty);
      await this.arrayEntityRequest({
        typesList: orderedSubTypes,
        ...baseParams
      });
    } else {
      const mappedTypes = dataResource.map(() => type);
      await this.arrayEntityRequest({
        typesList: mappedTypes,
        ...baseParams
      });
    }
  }
  async arrayEntityRequest({
    typesList,
    dataResourceList,
    incomingPath,
    dataAccessorName,
    entityDecoratorOptions,
    parentData
  }) {
    const resolveDataPromises = [];
    for (const [idx, type] of typesList.entries()) {
      const resourceEntity = dataResourceList[idx];
      const dataAccessor = `${dataAccessorName}[${idx}]`;
      const shouldResolve = await this.conditionalResolver({
        entityDecoratorOptions,
        meta: resourceEntity.meta,
        parentData: parentData,
        dataAccessorName: dataAccessor
      });
      if (!shouldResolve) continue;
      const entity = this.createEntity(resourceEntity);
      entity.path = `${incomingPath}${dataAccessor}`;
      this.entityMap.set(entity.path, entity);
      resolveDataPromises.push(this.httpClient.get(`/${entity.entityTypeId}/${entity.bundleId}/${entity.id}`).then(data => {
        const retrievedEntity = this.entityMap.get(entity.path);
        if (!retrievedEntity) throw new Error("Error retrieving entity");
        retrievedEntity.data = data;
        return {
          type,
          entity: retrievedEntity
        };
      }));
    }
    const resolvedData = await Promise.all(resolveDataPromises);
    const concurrentPromises = [];
    for (const item of resolvedData) {
      if (!item) continue;
      const {
        entity,
        type
      } = item;
      if (!entity.data) throw new Error("Entity has no resolved data");
      concurrentPromises.push(this.execute({
        target: type,
        data: entity.data,
        path: entity.path
      }));
    }
    await Promise.all(concurrentPromises);
  }
  createEntity(dataResource) {
    return new Entity({
      type: dataResource.type,
      id: dataResource.id,
      meta: dataResource.meta
    });
  }
}

/* eslint-disable @typescript-eslint/no-var-requires */
const set = require("lodash/set");
class EntityResolverService {
  constructor(httpClient) {
    this.httpClient = void 0;
    this.httpClient = httpClient;
  }
  async get(target, data, options) {
    const entityResolver = new EntityResolver(this.httpClient, options);
    const entityMap = await entityResolver.execute({
      target,
      data,
      options,
      path: ""
    });
    return classTransformer.plainToInstance(target, this.rebuildResolvedData(entityMap, data) ?? {}, {
      excludeExtraneousValues: true
    });
  }
  rebuildResolvedData(entityMap, data) {
    let rebuiltData = null;
    for (const entity of entityMap.values()) {
      rebuiltData = set(data, entity.path, {
        ...entity.data,
        _meta: entity.meta
      });
    }
    return rebuiltData;
  }
}

exports.Union = entity.Union;
exports.EntityResolverService = EntityResolverService;
exports.WithMeta = WithMeta;
